/* CSS for left navigation bar */

.ltx_page_navbar {
    display:block!important; position:fixed; left:80%; top:0px; width:20%;
    margin:0em; padding:1em; font: bold 75% sans-serif; overflow-y: auto; max-height: 95vh; }
.ltx_page_navbar ul {
    margin-left:-2em; }
.ltx_page_main {
    margin:0px; padding:1em 3em 1em 2em;
    width:75%; }
